#include "bila.cuh"



#include "cuda_runtime.h"
#include "device_launch_parameters.h"

#include <stdio.h>

#define threadX 16
#define threadY 16
#define threadZ 1

__global__ void bilakernel(uchar3 * src, uchar3* dst, int width, int height, int r, double sigmaC, double sigmaS)
{
	int x = threadIdx.x + blockDim.x * blockIdx.x;
	int y = threadIdx.y + blockDim.y * blockIdx.y;

	if (x >= width || y >= height)return;
	double mask = 0;
	double3 result;

	// mask.x = 0;
	// mask.y = 0;
	// mask.z = 0;
	result.x = 0;
	result.y = 0;
	result.z = 0;

	uchar3 origin = src[x + y * width];
	
	for(int i = -r; i<=r;i++ )
	{
		for(int j = -r; j<=r;j++)
		{
			if(x+i>=0&&x+i<width&&y+j>=0&&y+j<height)
			{
				uchar3 tmp = src[x + i + (y + j) * width];
				int3 sub;
				sub.x = tmp.x - origin.x;
				sub.y = tmp.y - origin.y;
				sub.z = tmp.z - origin.z;
				double masktmp = (exp(-0.5 / sigmaC / sigmaC * sub.x * sub.x) +
					exp(-0.5 / sigmaC / sigmaC * sub.y * sub.y) +
					exp(-0.5 / sigmaC / sigmaC * sub.z * sub.z)) *
					exp(-0.5 / sigmaS / sigmaS * (i * i + j * j));
				mask += masktmp;
				result.x += tmp.x * masktmp;
				result.y += tmp.y * masktmp;
				result.z += tmp.z * masktmp;
			}
		}
	}

	if(mask>0)
	{
		dst[x + y * width].x = result.x / mask;
		dst[x + y * width].y = result.y / mask;
		dst[x + y * width].z = result.z / mask;
	}
	else
	{
		dst[x + y * width].x = 0;
		dst[x + y * width].y = 0;
		dst[x + y * width].z = 0;
	}
	// if(x == 0 && y == 0)
	// {
	// 	printf("\n\n\n\n\n\n\n\n/////////////// on gpu  /////////////////// \n\n grid dim = (%d, %d) blockdim = (%d,%d)\n\n\n\n\n",gridDim.x,gridDim.y,blockDim.x,blockDim.y);
	// }
}


__global__ void bilakernel_v2(uchar3 *  src, uchar3* dst,int width,int height, int r, double sigmaC,double sigmaD )
{
	
	int x = threadIdx.x + blockDim.x * blockIdx.x;
	int y = threadIdx.y + blockDim.y * blockIdx.y;
	 
	
	
	if (x >= width || y >= height)return;
	const int blocksize = (2 * r + threadX) * (2 * r + threadY);
	const int id = x + y * width;
	extern  __shared__ uchar3 box[];// (2 * threadX)* (2 * threadY)];
	
	
	int i = 0;
	while(i* blockDim.x*blockDim.y+threadIdx.x+threadIdx.y*blockDim.x<blocksize)
	{
		int localindex = i * blockDim.x * blockDim.y + threadIdx.x + threadIdx.y * blockDim.x;
		
	
	
		i++;
	}

	for(int i = 0; i< 999999;i++)
	{
		int localIndex = threadIdx.x + threadIdx.y * blockDim.x + i * blockDim.x * blockDim.y;

		if(localIndex<(2*r+blockDim.x)*(2*r+blockDim.y))
		{
			int x_temp = localIndex % (blockDim.x + 2 * r);
			int y_temp = localIndex / (blockDim.x + 2 * r);
			// if(x_temp-r+blockDim.x*blockIdx.x<0 || y_temp-r+blockDim.y*blockIdx.y||
				// x_temp+r+blockDim.x*(blockIdx.x+1)-1>width)
			int x_real = x_temp - r + blockDim.x*blockIdx.x;
			int y_real = y_temp - r + blockDim.y * blockIdx.y;
			if(x_real<0||x_real>=width||y_real<0||y_real>=height)
			{
				box[localIndex].x = 0;
				box[localIndex].y = 0;
				box[localIndex].z = 0;
			}
			else
			{
				box[localIndex] = src[x_real + y_real * width];
				// printf("id = %d, localid  = %d, dstid = %d\n", id, localIndex, x_temp + y_temp * width);
				// if (blockIdx.x == 0 && blockIdx.y == 0) {
 				// }
			}
		}
		else
		{
			break;
		}
	}
	__syncthreads();


	// printf("src[%d] = [%d,%d,%d],localindex = %d boxcenter = [%d,%d,%d]\n", id,localIndex, src[id].x, src[id].y, src[id].z, box[r+r * (2*r + blockDim.x)].x, box[r + r * (2 * r + blockDim.x)].y, box[r + r * (2 * r + blockDim.x)].z);


	double coeff = 0;
	double sum[3] = { 0 };
	int idCenter = r + threadIdx.x + (r+threadIdx.y) * ( r*2 + blockDim.y);
	for(int i = -r; i<=r;i++)
	{
		for(int j = -r;j<=r;j++)
		{
			if(x+i>=0&&x+i<width&&y+j>=0&&y+j<height)
			{
				int idCur = r + i + threadIdx.x + (r + j + threadIdx.y) * (r * 2 + blockDim.y);
				double tmp = (
					exp(-0.5 / sigmaC / sigmaC * (box[idCur].x - box[idCenter].x) * (box[idCur].x - box[idCenter].x)) +
					exp(-0.5 / sigmaC / sigmaC * (box[idCur].y - box[idCenter].y) * (box[idCur].y - box[idCenter].y)) +
					exp(-0.5 / sigmaC / sigmaC * (box[idCur].z - box[idCenter].z) * (box[idCur].z - box[idCenter].z))
					) *
					exp(-0.5 / sigmaD / sigmaD *
						(i * i + j * j));
				coeff += tmp;
				sum[0] += tmp * box[idCur].x;
				sum[1] += tmp * box[idCur].y;
				sum[2] += tmp * box[idCur].z;
			}
	
		}
	}
	// printf("sum[%d] = [%f,%f,%f] coeff = %f\n", id,sum[0], sum[1], sum[2],coeff);
	if (coeff > 0) {
		dst[id].x = (uchar)(sum[0] / coeff);
		dst[id].y = (uchar)(sum[1] / coeff);
		dst[id].z = (uchar)(sum[2] / coeff);
	}
	else
	{
		dst[id].x = 0;
		dst[id].y = 0;
		dst[id].z = 0;
	}

	// printf("dst[%d] = [%d,%d,%d]\n", id, dst[id].x, dst[id].y, dst[id].z);
	// if (x == 50 && y == 50)
	// {
	// 	printf("\n\n\n\n\n\n\n\n/////////////// on gpu  /////////////////// \n\n grid dim = (%d, %d) blockdim = (%d,%d)\n\n\n\n\n", gridDim.x, gridDim.y, blockDim.x, blockDim.y);
	// 	printf("////box index///\n\n box[%d,%d] = %d,%d,%d\n\n\n", x, y, box[r * (2 * r + blockDim.x)].x, box[r * (2 * r + blockDim.x)].y, box[r * (2 * r + blockDim.x)].z);
	// }

}


void  mycu::bilateralFilter(cv::Mat src, cv::Mat dst, int r, double sigmaC, double sigmaS)
{
	// double sigmaCC = exp(-0.5 / sigmaC / sigmaC);
	// double sigmaSS = exp(-0.5 / sigmaS / sigmaS);

	if(src.type()!=CV_8UC3)
	{
		return;
	}
	int width = src.cols;
	int height = src.rows;

	uchar3* srcGpu;
	uchar3* dstGpu;

	cudaMalloc(&srcGpu, 3 * width * height);
	cudaMalloc(&dstGpu, 3 * width * height);

	cudaMemcpy(srcGpu, src.data, 3 * width * height, cudaMemcpyHostToDevice);

	// bilakernel << <dim3((width - 1) / threadX + 1, (height - 1) / threadY + 1), dim3(threadX, threadY) >> > (srcGpu, dstGpu, width, height, r, sigmaC, sigmaS);
	bilakernel_v2 <<<dim3((width - 1) / threadX + 1, (height - 1) / threadY + 1), dim3(threadX, threadY),(threadX+2*r)*(threadY+2*r)*4 >>> (srcGpu, dstGpu, width, height, r, sigmaC, sigmaS);

	cudaDeviceSynchronize();
	cudaError_t err = cudaGetLastError();
	std::cout << "err code = " << err << ", msg = " << cudaGetErrorString(err);
	cudaMemcpy(dst.data, dstGpu, 3 * width * height, cudaMemcpyDeviceToHost);

	cudaFree(srcGpu);
	cudaFree(dstGpu);
}
